# Copyright (c) 2011 OpenStack Foundation
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Glance"""

import copy

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy

from glance.common import exception
import glance.domain.proxy
from glance.i18n import _


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

DEFAULT_RULES = policy.Rules.from_dict({
    'context_is_admin': 'role:admin',
    'default': '@',
    'manage_image_cache': 'role:admin',
})


class Enforcer(policy.Enforcer):
    """Responsible for loading and enforcing rules"""

    def __init__(self):
        if CONF.find_file(CONF.oslo_policy.policy_file):
            kwargs = dict(rules=None, use_conf=True)
        else:
            kwargs = dict(rules=DEFAULT_RULES, use_conf=False)
        super(Enforcer, self).__init__(CONF, overwrite=False, **kwargs)

    def add_rules(self, rules):
        """Add new rules to the Rules object"""
        self.set_rules(rules, overwrite=False, use_conf=self.use_conf)

    def enforce(self, context, action, target):
        """Verifies that the action is valid on the target in this context.

           :param context: Glance request context
           :param action: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :raises: `glance.common.exception.Forbidden`
           :returns: A non-False value if access is allowed.
        """
        return super(Enforcer, self).enforce(action, target,
                                             context.to_policy_values(),
                                             do_raise=True,
                                             exc=exception.Forbidden,
                                             action=action)

    def check(self, context, action, target):
        """Verifies that the action is valid on the target in this context.

           :param context: Glance request context
           :param action: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :returns: A non-False value if access is allowed.
        """
        return super(Enforcer, self).enforce(action,
                                             target,
                                             context.to_policy_values())

    def check_is_admin(self, context):
        """Check if the given context is associated with an admin role,
           as defined via the 'context_is_admin' RBAC rule.

           :param context: Glance request context
           :returns: A non-False value if context role is admin.
        """
        return self.check(context, 'context_is_admin', context.to_dict())


class ImageRepoProxy(glance.domain.proxy.Repo):

    def __init__(self, image_repo, context, policy):
        self.context = context
        self.policy = policy
        self.image_repo = image_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(ImageRepoProxy, self).__init__(image_repo,
                                             item_proxy_class=ImageProxy,
                                             item_proxy_kwargs=proxy_kwargs)

    def get(self, image_id):
        try:
            image = super(ImageRepoProxy, self).get(image_id)
        except exception.NotFound:
            self.policy.enforce(self.context, 'get_image', {})
            raise
        else:
            self.policy.enforce(self.context, 'get_image', ImageTarget(image))
        return image

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_images', {})
        return super(ImageRepoProxy, self).list(*args, **kwargs)

    def save(self, image, from_state=None):
        self.policy.enforce(self.context, 'modify_image', image.target)
        return super(ImageRepoProxy, self).save(image, from_state=from_state)

    def add(self, image):
        self.policy.enforce(self.context, 'add_image', image.target)
        return super(ImageRepoProxy, self).add(image)


class ImageProxy(glance.domain.proxy.Image):

    def __init__(self, image, context, policy):
        self.image = image
        self.target = ImageTarget(image)
        self.context = context
        self.policy = policy
        super(ImageProxy, self).__init__(image)

    @property
    def visibility(self):
        return self.image.visibility

    @visibility.setter
    def visibility(self, value):
        if value == 'public':
            self.policy.enforce(self.context, 'publicize_image', self.target)
        self.image.visibility = value

    @property
    def locations(self):
        return ImageLocationsProxy(self.image.locations,
                                   self.context, self.policy)

    @locations.setter
    def locations(self, value):
        if not isinstance(value, (list, ImageLocationsProxy)):
            raise exception.Invalid(_('Invalid locations: %s') % value)
        self.policy.enforce(self.context, 'set_image_location', self.target)
        new_locations = list(value)
        if (set([loc['url'] for loc in self.image.locations]) -
                set([loc['url'] for loc in new_locations])):
            self.policy.enforce(self.context, 'delete_image_location',
                                self.target)
        self.image.locations = new_locations

    def delete(self):
        self.policy.enforce(self.context, 'delete_image', self.target)
        return self.image.delete()

    def deactivate(self):
        LOG.debug('Attempting deactivate')
        target = ImageTarget(self.image)
        self.policy.enforce(self.context, 'deactivate', target=target)
        LOG.debug('Deactivate allowed, continue')
        self.image.deactivate()

    def reactivate(self):
        LOG.debug('Attempting reactivate')
        target = ImageTarget(self.image)
        self.policy.enforce(self.context, 'reactivate', target=target)
        LOG.debug('Reactivate allowed, continue')
        self.image.reactivate()

    def get_data(self, *args, **kwargs):
        self.policy.enforce(self.context, 'download_image', self.target)
        return self.image.get_data(*args, **kwargs)

    def set_data(self, *args, **kwargs):
        self.policy.enforce(self.context, 'upload_image', self.target)
        return self.image.set_data(*args, **kwargs)


class ImageMemberProxy(glance.domain.proxy.ImageMember):

    def __init__(self, image_member, context, policy):
        super(ImageMemberProxy, self).__init__(image_member)
        self.image_member = image_member
        self.context = context
        self.policy = policy


class ImageFactoryProxy(glance.domain.proxy.ImageFactory):

    def __init__(self, image_factory, context, policy):
        self.image_factory = image_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(ImageFactoryProxy, self).__init__(image_factory,
                                                proxy_class=ImageProxy,
                                                proxy_kwargs=proxy_kwargs)

    def new_image(self, **kwargs):
        if kwargs.get('visibility') == 'public':
            self.policy.enforce(self.context, 'publicize_image', {})
        return super(ImageFactoryProxy, self).new_image(**kwargs)


class ImageMemberFactoryProxy(glance.domain.proxy.ImageMembershipFactory):

    def __init__(self, member_factory, context, policy):
        super(ImageMemberFactoryProxy, self).__init__(
            member_factory,
            proxy_class=ImageMemberProxy,
            proxy_kwargs={'context': context, 'policy': policy})


class ImageMemberRepoProxy(glance.domain.proxy.Repo):

    def __init__(self, member_repo, image, context, policy):
        self.member_repo = member_repo
        self.image = image
        self.target = ImageTarget(image)
        self.context = context
        self.policy = policy

    def add(self, member):
        self.policy.enforce(self.context, 'add_member', self.target)
        self.member_repo.add(member)

    def get(self, member_id):
        self.policy.enforce(self.context, 'get_member', self.target)
        return self.member_repo.get(member_id)

    def save(self, member, from_state=None):
        self.policy.enforce(self.context, 'modify_member', self.target)
        self.member_repo.save(member, from_state=from_state)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_members', self.target)
        return self.member_repo.list(*args, **kwargs)

    def remove(self, member):
        self.policy.enforce(self.context, 'delete_member', self.target)
        self.member_repo.remove(member)


class ImageLocationsProxy(object):

    __hash__ = None

    def __init__(self, locations, context, policy):
        self.locations = locations
        self.context = context
        self.policy = policy

    def __copy__(self):
        return type(self)(self.locations, self.context, self.policy)

    def __deepcopy__(self, memo):
        # NOTE(zhiyan): Only copy location entries, others can be reused.
        return type(self)(copy.deepcopy(self.locations, memo),
                          self.context, self.policy)

    def _get_checker(action, func_name):
        def _checker(self, *args, **kwargs):
            self.policy.enforce(self.context, action, {})
            method = getattr(self.locations, func_name)
            return method(*args, **kwargs)
        return _checker

    count = _get_checker('get_image_location', 'count')
    index = _get_checker('get_image_location', 'index')
    __getitem__ = _get_checker('get_image_location', '__getitem__')
    __contains__ = _get_checker('get_image_location', '__contains__')
    __len__ = _get_checker('get_image_location', '__len__')
    __cast = _get_checker('get_image_location', '__cast')
    __cmp__ = _get_checker('get_image_location', '__cmp__')
    __iter__ = _get_checker('get_image_location', '__iter__')

    append = _get_checker('set_image_location', 'append')
    extend = _get_checker('set_image_location', 'extend')
    insert = _get_checker('set_image_location', 'insert')
    reverse = _get_checker('set_image_location', 'reverse')
    __iadd__ = _get_checker('set_image_location', '__iadd__')
    __setitem__ = _get_checker('set_image_location', '__setitem__')

    pop = _get_checker('delete_image_location', 'pop')
    remove = _get_checker('delete_image_location', 'remove')
    __delitem__ = _get_checker('delete_image_location', '__delitem__')
    __delslice__ = _get_checker('delete_image_location', '__delslice__')

    del _get_checker


class TaskProxy(glance.domain.proxy.Task):

    def __init__(self, task, context, policy):
        self.task = task
        self.context = context
        self.policy = policy
        super(TaskProxy, self).__init__(task)


class TaskStubProxy(glance.domain.proxy.TaskStub):

    def __init__(self, task_stub, context, policy):
        self.task_stub = task_stub
        self.context = context
        self.policy = policy
        super(TaskStubProxy, self).__init__(task_stub)


class TaskRepoProxy(glance.domain.proxy.TaskRepo):

    def __init__(self, task_repo, context, task_policy):
        self.context = context
        self.policy = task_policy
        self.task_repo = task_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(TaskRepoProxy,
              self).__init__(task_repo,
                             task_proxy_class=TaskProxy,
                             task_proxy_kwargs=proxy_kwargs)

    def get(self, task_id):
        self.policy.enforce(self.context, 'get_task', {})
        return super(TaskRepoProxy, self).get(task_id)

    def add(self, task):
        self.policy.enforce(self.context, 'add_task', {})
        super(TaskRepoProxy, self).add(task)

    def save(self, task):
        self.policy.enforce(self.context, 'modify_task', {})
        super(TaskRepoProxy, self).save(task)


class TaskStubRepoProxy(glance.domain.proxy.TaskStubRepo):

    def __init__(self, task_stub_repo, context, task_policy):
        self.context = context
        self.policy = task_policy
        self.task_stub_repo = task_stub_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(TaskStubRepoProxy,
              self).__init__(task_stub_repo,
                             task_stub_proxy_class=TaskStubProxy,
                             task_stub_proxy_kwargs=proxy_kwargs)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_tasks', {})
        return super(TaskStubRepoProxy, self).list(*args, **kwargs)


class TaskFactoryProxy(glance.domain.proxy.TaskFactory):

    def __init__(self, task_factory, context, policy):
        self.task_factory = task_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(TaskFactoryProxy, self).__init__(
            task_factory,
            task_proxy_class=TaskProxy,
            task_proxy_kwargs=proxy_kwargs)


class ImageTarget(object):
    SENTINEL = object()

    def __init__(self, target):
        """Initialize the object

        :param target: Object being targeted
        """
        self.target = target

    def __getitem__(self, key):
        """Return the value of 'key' from the target.

        If the target has the attribute 'key', return it.

        :param key: value to retrieve
        """
        key = self.key_transforms(key)

        value = getattr(self.target, key, self.SENTINEL)
        if value is self.SENTINEL:
            extra_properties = getattr(self.target, 'extra_properties', None)
            if extra_properties is not None:
                value = extra_properties[key]
            else:
                value = None
        return value

    def key_transforms(self, key):
        if key == 'id':
            key = 'image_id'

        return key


# Metadef Namespace classes
class MetadefNamespaceProxy(glance.domain.proxy.MetadefNamespace):

    def __init__(self, namespace, context, policy):
        self.namespace_input = namespace
        self.context = context
        self.policy = policy
        super(MetadefNamespaceProxy, self).__init__(namespace)


class MetadefNamespaceRepoProxy(glance.domain.proxy.MetadefNamespaceRepo):

    def __init__(self, namespace_repo, context, namespace_policy):
        self.context = context
        self.policy = namespace_policy
        self.namespace_repo = namespace_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefNamespaceRepoProxy,
              self).__init__(namespace_repo,
                             namespace_proxy_class=MetadefNamespaceProxy,
                             namespace_proxy_kwargs=proxy_kwargs)

    def get(self, namespace):
        self.policy.enforce(self.context, 'get_metadef_namespace', {})
        return super(MetadefNamespaceRepoProxy, self).get(namespace)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_metadef_namespaces', {})
        return super(MetadefNamespaceRepoProxy, self).list(*args, **kwargs)

    def save(self, namespace):
        self.policy.enforce(self.context, 'modify_metadef_namespace', {})
        return super(MetadefNamespaceRepoProxy, self).save(namespace)

    def add(self, namespace):
        self.policy.enforce(self.context, 'add_metadef_namespace', {})
        return super(MetadefNamespaceRepoProxy, self).add(namespace)


class MetadefNamespaceFactoryProxy(
        glance.domain.proxy.MetadefNamespaceFactory):

    def __init__(self, meta_namespace_factory, context, policy):
        self.meta_namespace_factory = meta_namespace_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefNamespaceFactoryProxy, self).__init__(
            meta_namespace_factory,
            meta_namespace_proxy_class=MetadefNamespaceProxy,
            meta_namespace_proxy_kwargs=proxy_kwargs)


# Metadef Object classes
class MetadefObjectProxy(glance.domain.proxy.MetadefObject):

    def __init__(self, meta_object, context, policy):
        self.meta_object = meta_object
        self.context = context
        self.policy = policy
        super(MetadefObjectProxy, self).__init__(meta_object)


class MetadefObjectRepoProxy(glance.domain.proxy.MetadefObjectRepo):

    def __init__(self, object_repo, context, object_policy):
        self.context = context
        self.policy = object_policy
        self.object_repo = object_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefObjectRepoProxy,
              self).__init__(object_repo,
                             object_proxy_class=MetadefObjectProxy,
                             object_proxy_kwargs=proxy_kwargs)

    def get(self, namespace, object_name):
        self.policy.enforce(self.context, 'get_metadef_object', {})
        return super(MetadefObjectRepoProxy, self).get(namespace, object_name)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_metadef_objects', {})
        return super(MetadefObjectRepoProxy, self).list(*args, **kwargs)

    def save(self, meta_object):
        self.policy.enforce(self.context, 'modify_metadef_object', {})
        return super(MetadefObjectRepoProxy, self).save(meta_object)

    def add(self, meta_object):
        self.policy.enforce(self.context, 'add_metadef_object', {})
        return super(MetadefObjectRepoProxy, self).add(meta_object)


class MetadefObjectFactoryProxy(glance.domain.proxy.MetadefObjectFactory):

    def __init__(self, meta_object_factory, context, policy):
        self.meta_object_factory = meta_object_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefObjectFactoryProxy, self).__init__(
            meta_object_factory,
            meta_object_proxy_class=MetadefObjectProxy,
            meta_object_proxy_kwargs=proxy_kwargs)


# Metadef ResourceType classes
class MetadefResourceTypeProxy(glance.domain.proxy.MetadefResourceType):

    def __init__(self, meta_resource_type, context, policy):
        self.meta_resource_type = meta_resource_type
        self.context = context
        self.policy = policy
        super(MetadefResourceTypeProxy, self).__init__(meta_resource_type)


class MetadefResourceTypeRepoProxy(
        glance.domain.proxy.MetadefResourceTypeRepo):

    def __init__(self, resource_type_repo, context, resource_type_policy):
        self.context = context
        self.policy = resource_type_policy
        self.resource_type_repo = resource_type_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefResourceTypeRepoProxy, self).__init__(
            resource_type_repo,
            resource_type_proxy_class=MetadefResourceTypeProxy,
            resource_type_proxy_kwargs=proxy_kwargs)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'list_metadef_resource_types', {})
        return super(MetadefResourceTypeRepoProxy, self).list(*args, **kwargs)

    def get(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_metadef_resource_type', {})
        return super(MetadefResourceTypeRepoProxy, self).get(*args, **kwargs)

    def add(self, resource_type):
        self.policy.enforce(self.context,
                            'add_metadef_resource_type_association', {})
        return super(MetadefResourceTypeRepoProxy, self).add(resource_type)


class MetadefResourceTypeFactoryProxy(
        glance.domain.proxy.MetadefResourceTypeFactory):

    def __init__(self, resource_type_factory, context, policy):
        self.resource_type_factory = resource_type_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefResourceTypeFactoryProxy, self).__init__(
            resource_type_factory,
            resource_type_proxy_class=MetadefResourceTypeProxy,
            resource_type_proxy_kwargs=proxy_kwargs)


# Metadef namespace properties classes
class MetadefPropertyProxy(glance.domain.proxy.MetadefProperty):

    def __init__(self, namespace_property, context, policy):
        self.namespace_property = namespace_property
        self.context = context
        self.policy = policy
        super(MetadefPropertyProxy, self).__init__(namespace_property)


class MetadefPropertyRepoProxy(glance.domain.proxy.MetadefPropertyRepo):

    def __init__(self, property_repo, context, object_policy):
        self.context = context
        self.policy = object_policy
        self.property_repo = property_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefPropertyRepoProxy, self).__init__(
            property_repo,
            property_proxy_class=MetadefPropertyProxy,
            property_proxy_kwargs=proxy_kwargs)

    def get(self, namespace, property_name):
        self.policy.enforce(self.context, 'get_metadef_property', {})
        return super(MetadefPropertyRepoProxy, self).get(namespace,
                                                         property_name)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_metadef_properties', {})
        return super(MetadefPropertyRepoProxy, self).list(
            *args, **kwargs)

    def save(self, namespace_property):
        self.policy.enforce(self.context, 'modify_metadef_property', {})
        return super(MetadefPropertyRepoProxy, self).save(
            namespace_property)

    def add(self, namespace_property):
        self.policy.enforce(self.context, 'add_metadef_property', {})
        return super(MetadefPropertyRepoProxy, self).add(
            namespace_property)


class MetadefPropertyFactoryProxy(glance.domain.proxy.MetadefPropertyFactory):

    def __init__(self, namespace_property_factory, context, policy):
        self.namespace_property_factory = namespace_property_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefPropertyFactoryProxy, self).__init__(
            namespace_property_factory,
            property_proxy_class=MetadefPropertyProxy,
            property_proxy_kwargs=proxy_kwargs)


# Metadef Tag classes
class MetadefTagProxy(glance.domain.proxy.MetadefTag):

    def __init__(self, meta_tag, context, policy):
        self.context = context
        self.policy = policy
        super(MetadefTagProxy, self).__init__(meta_tag)


class MetadefTagRepoProxy(glance.domain.proxy.MetadefTagRepo):

    def __init__(self, tag_repo, context, tag_policy):
        self.context = context
        self.policy = tag_policy
        self.tag_repo = tag_repo
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefTagRepoProxy,
              self).__init__(tag_repo,
                             tag_proxy_class=MetadefTagProxy,
                             tag_proxy_kwargs=proxy_kwargs)

    def get(self, namespace, tag_name):
        self.policy.enforce(self.context, 'get_metadef_tag', {})
        return super(MetadefTagRepoProxy, self).get(namespace, tag_name)

    def list(self, *args, **kwargs):
        self.policy.enforce(self.context, 'get_metadef_tags', {})
        return super(MetadefTagRepoProxy, self).list(*args, **kwargs)

    def save(self, meta_tag):
        self.policy.enforce(self.context, 'modify_metadef_tag', {})
        return super(MetadefTagRepoProxy, self).save(meta_tag)

    def add(self, meta_tag):
        self.policy.enforce(self.context, 'add_metadef_tag', {})
        return super(MetadefTagRepoProxy, self).add(meta_tag)

    def add_tags(self, meta_tags):
        self.policy.enforce(self.context, 'add_metadef_tags', {})
        return super(MetadefTagRepoProxy, self).add_tags(meta_tags)


class MetadefTagFactoryProxy(glance.domain.proxy.MetadefTagFactory):

    def __init__(self, meta_tag_factory, context, policy):
        self.meta_tag_factory = meta_tag_factory
        self.context = context
        self.policy = policy
        proxy_kwargs = {'context': self.context, 'policy': self.policy}
        super(MetadefTagFactoryProxy, self).__init__(
            meta_tag_factory,
            meta_tag_proxy_class=MetadefTagProxy,
            meta_tag_proxy_kwargs=proxy_kwargs)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import uuid

from cursive import exception as cursive_exception
import glance_store
import mock
import six
import webob

import glance.api.policy
import glance.api.v2.image_data
from glance.common import exception
from glance.common import wsgi
from glance.tests.unit import base
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils


class Raise(object):
    def __init__(self, exc):
        self.exc = exc

    def __call__(self, *args, **kwargs):
        raise self.exc


class FakeImage(object):
    def __init__(self, image_id=None, data=None, checksum=None, size=0,
                 virtual_size=0, locations=None, container_format='bear',
                 disk_format='rawr', status=None):
        self.image_id = image_id
        self.data = data
        self.checksum = checksum
        self.size = size
        self.virtual_size = virtual_size
        self.locations = locations
        self.container_format = container_format
        self.disk_format = disk_format
        self._status = status

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if isinstance(self._status, BaseException):
            raise self._status
        else:
            self._status = value

    def get_data(self, *args, **kwargs):
        return self.data

    def set_data(self, data, size=None):
        self.data = ''.join(data)
        self.size = size
        self.status = 'modified-by-fake'


class FakeImageRepo(object):
    def __init__(self, result=None):
        self.result = result

    def get(self, image_id):
        if isinstance(self.result, BaseException):
            raise self.result
        else:
            return self.result

    def save(self, image, from_state=None):
        self.saved_image = image


class FakeGateway(object):
    def __init__(self, repo):
        self.repo = repo

    def get_repo(self, context):
        return self.repo


class TestImagesController(base.StoreClearingUnitTest):
    def setUp(self):
        super(TestImagesController, self).setUp()

        self.config(debug=True)
        self.image_repo = FakeImageRepo()
        self.gateway = FakeGateway(self.image_repo)
        self.controller = glance.api.v2.image_data.ImageDataController(
            gateway=self.gateway)

    def test_download(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd',
                          locations=[{'url': 'http://example.com/image',
                                      'metadata': {}, 'status': 'active'}])
        self.image_repo.result = image
        image = self.controller.download(request, unit_test_utils.UUID1)
        self.assertEqual('abcd', image.image_id)

    def test_download_deactivated(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd',
                          status='deactivated',
                          locations=[{'url': 'http://example.com/image',
                                      'metadata': {}, 'status': 'active'}])
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_no_location(self):
        # NOTE(mclaren): NoContent will be raised by the ResponseSerializer
        # That's tested below.
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = FakeImage('abcd')
        image = self.controller.download(request, unit_test_utils.UUID2)
        self.assertEqual('abcd', image.image_id)

    def test_download_non_existent_image(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.NotFound()
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_forbidden(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.Forbidden()
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_ok_when_get_image_location_forbidden(self):
        class ImageLocations(object):
            def __len__(self):
                raise exception.Forbidden()

        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        self.image_repo.result = image
        image.locations = ImageLocations()
        image = self.controller.download(request, unit_test_utils.UUID1)
        self.assertEqual('abcd', image.image_id)

    def test_upload(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)

    def test_upload_status(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        self.image_repo.result = image
        insurance = {'called': False}

        def read_data():
            insurance['called'] = True
            self.assertEqual('saving', self.image_repo.saved_image.status)
            yield 'YYYY'

        self.controller.upload(request, unit_test_utils.UUID2,
                               read_data(), None)
        self.assertTrue(insurance['called'])
        self.assertEqual('modified-by-fake',
                         self.image_repo.saved_image.status)

    def test_upload_no_size(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', None)
        self.assertEqual('YYYY', image.data)
        self.assertIsNone(image.size)

    def test_upload_invalid(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        image.status = ValueError()
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)

    def test_upload_with_expired_token(self):
        def side_effect(image, from_state=None):
            if from_state == 'saving':
                raise exception.NotAuthenticated()

        mocked_save = mock.Mock(side_effect=side_effect)
        mocked_delete = mock.Mock()
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        image.delete = mocked_delete
        self.image_repo.result = image
        self.image_repo.save = mocked_save
        self.assertRaises(webob.exc.HTTPUnauthorized, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)
        self.assertEqual(3, mocked_save.call_count)
        mocked_delete.assert_called_once_with()

    def test_upload_non_existent_image_during_save_initiates_deletion(self):
        def fake_save_not_found(self, from_state=None):
            raise exception.ImageNotFound()

        def fake_save_conflict(self, from_state=None):
            raise exception.Conflict()

        for fun in [fake_save_not_found, fake_save_conflict]:
            request = unit_test_utils.get_fake_request()
            image = FakeImage('abcd', locations=['http://example.com/image'])
            self.image_repo.result = image
            self.image_repo.save = fun
            image.delete = mock.Mock()
            self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                              request, str(uuid.uuid4()), 'ABC', 3)
            self.assertTrue(image.delete.called)

    def test_upload_non_existent_image_raises_image_not_found_exception(self):
        def fake_save(self, from_state=None):
            raise exception.ImageNotFound()

        def fake_delete():
            raise exception.ImageNotFound()

        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd', locations=['http://example.com/image'])
        self.image_repo.result = image
        self.image_repo.save = fake_save
        image.delete = fake_delete
        self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_non_existent_image_raises_store_not_found_exception(self):
        def fake_save(self, from_state=None):
            raise glance_store.NotFound()

        def fake_delete():
            raise exception.ImageNotFound()

        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd', locations=['http://example.com/image'])
        self.image_repo.result = image
        self.image_repo.save = fake_save
        image.delete = fake_delete
        self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_non_existent_image_before_save(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.NotFound()
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_data_exists(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage()
        exc = exception.InvalidImageStatusTransition(cur_status='active',
                                                     new_status='queued')
        image.set_data = Raise(exc)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPConflict, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)

    def test_upload_storage_full(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage()
        image.set_data = Raise(glance_store.StorageFull)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YYYYYYY', 7)

    def test_upload_signature_verification_fails(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage()
        image.set_data = Raise(cursive_exception.SignatureVerificationError)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)
        self.assertEqual('killed', self.image_repo.saved_image.status)

    def test_image_size_limit_exceeded(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage()
        image.set_data = Raise(exception.ImageSizeLimitExceeded)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYYYYY', 7)

    def test_upload_storage_quota_full(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.StorageQuotaFull("message")
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYYYYY', 7)

    def test_upload_storage_forbidden(self):
        request = unit_test_utils.get_fake_request(user=unit_test_utils.USER2)
        image = FakeImage()
        image.set_data = Raise(exception.Forbidden)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    def test_upload_storage_internal_error(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.ServerError()
        self.assertRaises(exception.ServerError,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'ABC', 3)

    def test_upload_storage_write_denied(self):
        request = unit_test_utils.get_fake_request(user=unit_test_utils.USER3)
        image = FakeImage()
        image.set_data = Raise(glance_store.StorageWriteDenied)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPServiceUnavailable,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    def test_upload_storage_store_disabled(self):
        """Test that uploading an image file raises StoreDisabled exception"""
        request = unit_test_utils.get_fake_request(user=unit_test_utils.USER3)
        image = FakeImage()
        image.set_data = Raise(glance_store.StoreAddDisabled)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPGone,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    @mock.patch("glance.common.trust_auth.TokenRefresher")
    def test_upload_with_trusts(self, mock_refresher):
        """Test that uploading with registry correctly uses trusts"""
        # initialize trust environment
        self.config(data_api='glance.db.registry.api')
        refresher = mock.MagicMock()
        mock_refresher.return_value = refresher
        refresher.refresh_token.return_value = "fake_token"
        # request an image upload
        request = unit_test_utils.get_fake_request()
        request.environ['keystone.token_auth'] = mock.MagicMock()
        request.environ['keystone.token_info'] = {
            'token': {
                'roles': [{'name': 'FakeRole', 'id': 'FakeID'}]
            }
        }
        image = FakeImage('abcd')
        self.image_repo.result = image
        mock_fake_save = mock.Mock()
        mock_fake_save.side_effect = [None, exception.NotAuthenticated, None]
        temp_save = FakeImageRepo.save
        # mocking save to raise NotAuthenticated on the second call
        FakeImageRepo.save = mock_fake_save
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        # check image data
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)
        FakeImageRepo.save = temp_save
        # check that token has been correctly acquired and deleted
        mock_refresher.assert_called_once_with(
            request.environ['keystone.token_auth'],
            request.context.tenant, ['FakeRole'])
        refresher.refresh_token.assert_called_once_with()
        refresher.release_resources.assert_called_once_with()
        self.assertEqual("fake_token", request.context.auth_token)

    @mock.patch("glance.common.trust_auth.TokenRefresher")
    def test_upload_with_trusts_fails(self, mock_refresher):
        """Test upload with registry if trust was not successfully created"""
        # initialize trust environment
        self.config(data_api='glance.db.registry.api')
        mock_refresher().side_effect = Exception()
        # request an image upload
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        # check image data
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)
        # check that the token has not been updated
        self.assertEqual(0, mock_refresher().refresh_token.call_count)

    def _test_upload_download_prepare_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        prepare_payload = output['meta'].copy()
        prepare_payload['checksum'] = None
        prepare_payload['size'] = None
        prepare_payload['virtual_size'] = None
        prepare_payload['location'] = None
        prepare_payload['status'] = 'queued'
        del prepare_payload['updated_at']
        prepare_log = {
            'notification_type': "INFO",
            'event_type': "image.prepare",
            'payload': prepare_payload,
        }
        self.assertEqual(3, len(output_log))
        prepare_updated_at = output_log[0]['payload']['updated_at']
        del output_log[0]['payload']['updated_at']
        self.assertLessEqual(prepare_updated_at, output['meta']['updated_at'])
        self.assertEqual(prepare_log, output_log[0])

    def _test_upload_download_upload_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        upload_payload = output['meta'].copy()
        upload_log = {
            'notification_type': "INFO",
            'event_type': "image.upload",
            'payload': upload_payload,
        }
        self.assertEqual(3, len(output_log))
        self.assertEqual(upload_log, output_log[1])

    def _test_upload_download_activate_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        activate_payload = output['meta'].copy()
        activate_log = {
            'notification_type': "INFO",
            'event_type': "image.activate",
            'payload': activate_payload,
        }
        self.assertEqual(3, len(output_log))
        self.assertEqual(activate_log, output_log[2])

    def test_restore_image_when_upload_failed(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('fake')
        image.set_data = Raise(glance_store.StorageWriteDenied)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPServiceUnavailable,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'ZZZ', 3)
        self.assertEqual('queued', self.image_repo.saved_image.status)


class TestImageDataDeserializer(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageDataDeserializer, self).setUp()
        self.deserializer = glance.api.v2.image_data.RequestDeserializer()

    def test_upload(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        request.body = b'YYY'
        request.headers['Content-Length'] = 3
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 3}
        self.assertEqual(expected, output)

    def test_upload_chunked(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        # If we use body_file, webob assumes we want to do a chunked upload,
        # ignoring the Content-Length header
        request.body_file = six.StringIO('YYY')
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual('YYY', data.read())
        expected = {'size': None}
        self.assertEqual(expected, output)

    def test_upload_chunked_with_content_length(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        request.body_file = six.BytesIO(b'YYY')
        # The deserializer shouldn't care if the Content-Length is
        # set when the user is attempting to send chunked data.
        request.headers['Content-Length'] = 3
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 3}
        self.assertEqual(expected, output)

    def test_upload_with_incorrect_content_length(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        # The deserializer shouldn't care if the Content-Length and
        # actual request body length differ. That job is left up
        # to the controller
        request.body = b'YYY'
        request.headers['Content-Length'] = 4
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 4}
        self.assertEqual(expected, output)

    def test_upload_wrong_content_type(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/json'
        request.body = b'YYYYY'
        self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                          self.deserializer.upload, request)

        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-st'
        request.body = b'YYYYY'
        self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                          self.deserializer.upload, request)


class TestImageDataSerializer(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageDataSerializer, self).setUp()
        self.serializer = glance.api.v2.image_data.ResponseSerializer()

    def test_download(self):
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
        self.serializer.download(response, image)
        self.assertEqual(b'ZZZ', response.body)
        self.assertEqual('3', response.headers['Content-Length'])
        self.assertNotIn('Content-MD5', response.headers)
        self.assertEqual('application/octet-stream',
                         response.headers['Content-Type'])

    def test_download_with_checksum(self):
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        checksum = '0745064918b49693cca64d6b6a13d28a'
        image = FakeImage(size=3, checksum=checksum, data=[b'Z', b'Z', b'Z'])
        self.serializer.download(response, image)
        self.assertEqual(b'ZZZ', response.body)
        self.assertEqual('3', response.headers['Content-Length'])
        self.assertEqual(checksum, response.headers['Content-MD5'])
        self.assertEqual('application/octet-stream',
                         response.headers['Content-Type'])

    def test_download_forbidden(self):
        """Make sure the serializer can return 403 forbidden error instead of
        500 internal server error.
        """
        def get_data(*args, **kwargs):
            raise exception.Forbidden()

        self.stubs.Set(glance.api.policy.ImageProxy,
                       'get_data',
                       get_data)
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=iter('ZZZ'))
        image.get_data = get_data
        self.assertRaises(webob.exc.HTTPForbidden,
                          self.serializer.download,
                          response, image)

    def test_download_no_content(self):
        """Test image download returns HTTPNoContent

        Make sure that serializer returns 204 no content error in case of
        image data is not available at specified location.
        """
        with mock.patch.object(glance.api.policy.ImageProxy,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.NotFound(image="image")

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPNoContent,
                              self.serializer.download,
                              response, image)

    def test_download_service_unavailable(self):
        """Test image download returns HTTPServiceUnavailable."""
        with mock.patch.object(glance.api.policy.ImageProxy,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.RemoteServiceUnavailable()

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPServiceUnavailable,
                              self.serializer.download,
                              response, image)

    def test_download_store_get_not_support(self):
        """Test image download returns HTTPBadRequest.

        Make sure that serializer returns 400 bad request error in case of
        getting images from this store is not supported at specified location.
        """
        with mock.patch.object(glance.api.policy.ImageProxy,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.StoreGetNotSupported()

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.serializer.download,
                              response, image)

    def test_download_store_random_get_not_support(self):
        """Test image download returns HTTPBadRequest.

        Make sure that serializer returns 400 bad request error in case of
        getting randomly images from this store is not supported at
        specified location.
        """
        with mock.patch.object(glance.api.policy.ImageProxy,
                               'get_data') as m_get_data:
            err = glance_store.StoreRandomGetNotSupported(offset=0,
                                                          chunk_size=0)
            m_get_data.side_effect = err

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = m_get_data
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.serializer.download,
                              response, image)

    def test_upload(self):
        request = webob.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        self.serializer.upload(response, {})
        self.assertEqual(204, response.status_int)
        self.assertEqual('0', response.headers['Content-Length'])



# Copyright (c) 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glance.common import exception
from glance.common.glare import definitions
import glance.contrib.plugins.image_artifact.v1_1.image as v1_1

# Since this is not in the test-requirements.txt and the class below,
# ImageAsAnArtifact, is pending removal a try except is added to prevent
# an ImportError when module docs are generated
try:
    import glanceclient
except ImportError:
    glanceclient = None


from glance.i18n import _


class ImageAsAnArtifact(v1_1.ImageAsAnArtifact):
    __type_version__ = '2.0'

    file = definitions.BinaryObject(required=False)
    legacy_image_id = definitions.String(required=False, mutable=False,
                                         pattern=R'[0-9a-f]{8}-[0-9a-f]{4}'
                                                 R'-4[0-9a-f]{3}-[89ab]'
                                                 R'[0-9a-f]{3}-[0-9a-f]{12}')

    def __pre_publish__(self, context, *args, **kwargs):
        super(ImageAsAnArtifact, self).__pre_publish__(*args, **kwargs)
        if self.file is None and self.legacy_image_id is None:
            raise exception.InvalidArtifactPropertyValue(
                message=_("Either a file or a legacy_image_id has to be "
                          "specified")
            )
        if self.file is not None and self.legacy_image_id is not None:
            raise exception.InvalidArtifactPropertyValue(
                message=_("Both file and legacy_image_id may not be "
                          "specified at the same time"))

        if self.legacy_image_id:
            glance_endpoint = next(service['endpoints'][0]['publicURL']
                                   for service in context.service_catalog
                                   if service['name'] == 'glance')
            # Ensure glanceclient is imported correctly since we are catching
            # the ImportError on initialization
            if glanceclient == None:
                raise ImportError(_("Glance client not installed"))

            try:
                client = glanceclient.Client(version=2,
                                             endpoint=glance_endpoint,
                                             token=context.auth_token)
                legacy_image = client.images.get(self.legacy_image_id)
            except Exception:
                raise exception.InvalidArtifactPropertyValue(
                    message=_('Unable to get legacy image')
                )
            if legacy_image is not None:
                self.file = definitions.Blob(size=legacy_image.size,
                                             locations=[
                                                 {
                                                     "status": "active",
                                                     "value":
                                                     legacy_image.direct_url
                                                 }],
                                             checksum=legacy_image.checksum,
                                             item_key=legacy_image.id)
            else:
                raise exception.InvalidArtifactPropertyValue(
                    message=_("Legacy image was not found")
                )



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import uuid

import mock
from oslo_serialization import jsonutils
import pkg_resources
import requests

from glance.api.glare.v0_1 import glare
from glance.api.glare.v0_1 import router
from glance.common.glare import definitions
from glance.common.glare import loader
from glance.common import wsgi
from glance.tests import functional


class Artifact(definitions.ArtifactType):
    __type_name__ = "WithProps"
    prop1 = definitions.String()
    prop2 = definitions.Integer()
    prop_list = definitions.Array(item_type=definitions.Integer())
    tuple_prop = definitions.Array(item_type=[definitions.Integer(),
                                              definitions.Boolean()])
    dict_prop = definitions.Dict(properties={
        "foo": definitions.String(),
        "bar_list": definitions.Array(definitions.Integer())})
    dict_prop_strval = definitions.Dict(properties=definitions.String())
    depends_on = definitions.ArtifactReference()
    depends_on_list = definitions.ArtifactReferenceList()


class ArtifactNoProps(definitions.ArtifactType):
    __type_name__ = "NoProp"


class ArtifactNoProps1(definitions.ArtifactType):
    __type_name__ = "NoProp"
    __type_version__ = "0.5"


class ArtifactWithBlob(definitions.ArtifactType):
    __type_name__ = "WithBlob"
    blob1 = definitions.BinaryObject()
    blob_list = definitions.BinaryObjectList()


def _create_resource():
    plugins = None
    mock_this = 'stevedore.extension.ExtensionManager._find_entry_points'
    with mock.patch(mock_this) as fep:
        path = 'glance.tests.functional.glare.test_glare'
        fep.return_value = [
            pkg_resources.EntryPoint.parse('WithProps=%s:Artifact' % path),
            pkg_resources.EntryPoint.parse(
                'NoProp=%s:ArtifactNoProps' % path),
            pkg_resources.EntryPoint.parse(
                'NoProp=%s:ArtifactNoProps1' % path),
            pkg_resources.EntryPoint.parse(
                'WithBlob=%s:ArtifactWithBlob' % path)
        ]
        plugins = loader.ArtifactsPluginLoader('glance.artifacts.types')
    deserializer = glare.RequestDeserializer(plugins=plugins)
    serializer = glare.ResponseSerializer()
    controller = glare.ArtifactsController(plugins=plugins)
    return wsgi.Resource(controller, deserializer, serializer)


class TestRouter(router.API):
    def _get_artifacts_resource(self):
        return _create_resource()


class TestArtifacts(functional.FunctionalTest):

    users = {
        'user1': {
            'id': str(uuid.uuid4()),
            'tenant_id': str(uuid.uuid4()),
            'token': str(uuid.uuid4()),
            'role': 'member'
        },
        'user2': {
            'id': str(uuid.uuid4()),
            'tenant_id': str(uuid.uuid4()),
            'token': str(uuid.uuid4()),
            'role': 'member'
        },
        'admin': {
            'id': str(uuid.uuid4()),
            'tenant_id': str(uuid.uuid4()),
            'token': str(uuid.uuid4()),
            'role': 'admin'
        }
    }

    def setUp(self):
        super(TestArtifacts, self).setUp()
        self._set_user('user1')
        self.api_server.server_name = 'glare'
        self.api_server.server_module = 'glance.cmd.glare'
        self.api_server.deployment_flavor = 'noauth'
        self.start_servers(**self.__dict__.copy())

    def tearDown(self):
        self.stop_servers()
        self._reset_database(self.api_server.sql_connection)
        super(TestArtifacts, self).tearDown()

    def _url(self, path):
        return 'http://127.0.0.1:%d/v0.1/artifacts%s' % (self.api_port, path)

    def _set_user(self, username):
        if username not in self.users:
            raise KeyError
        self.current_user = username

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': self.users[self.current_user]['token'],
            'X-User-Id': self.users[self.current_user]['id'],
            'X-Tenant-Id': self.users[self.current_user]['tenant_id'],
            'X-Roles': self.users[self.current_user]['role'],
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def start_servers(self, **kwargs):
        # noqa
        new_paste_conf_base = """[pipeline:glare-api]
pipeline = versionnegotiation unauthenticated-context rootapp

[pipeline:glare-api-fakeauth]
pipeline = versionnegotiation fakeauth context rootapp

[pipeline:glare-api-noauth]
pipeline = versionnegotiation context rootapp

[composite:rootapp]
paste.composite_factory = glance.api:root_app_factory
/: apiversions
/v0.1: glareapi

[app:apiversions]
paste.app_factory = glance.api.glare.versions:create_resource

[app:glareapi]
paste.app_factory =
   glance.tests.functional.glare.test_glare:TestRouter.factory

[filter:versionnegotiation]
paste.filter_factory =
 glance.api.middleware.version_negotiation:
   GlareVersionNegotiationFilter.factory

[filter:context]
paste.filter_factory = glance.api.middleware.context:ContextMiddleware.factory

[filter:unauthenticated-context]
paste.filter_factory =
 glance.api.middleware.context:UnauthenticatedContextMiddleware.factory

[filter:fakeauth]
paste.filter_factory = glance.tests.utils:FakeAuthMiddleware.factory
"""
        self.cleanup()
        self.api_server.paste_conf_base = new_paste_conf_base
        super(TestArtifacts, self).start_servers(**kwargs)

    def _create_artifact(self, type_name, type_version='1.0', data=None,
                         status=201):
        # create an artifact first
        artifact_data = data or {'name': 'artifact-1',
                                 'version': '12'}
        return self._check_artifact_post('/%s/v%s/drafts' % (type_name,
                                                             type_version),
                                         artifact_data, status=status)

    def _check_artifact_method(self, method, url, data=None, status=200,
                               headers=None):
        if not headers:
            headers = self._headers()
        else:
            headers = self._headers(headers)
        headers.setdefault("Content-Type", "application/json")
        if 'application/json' in headers['Content-Type']:
            data = jsonutils.dumps(data)
        response = getattr(requests, method)(self._url(url), headers=headers,
                                             data=data)
        self.assertEqual(status, response.status_code)
        if status >= 400:
            return response.text
        if "application/json" in response.headers["content-type"]:
            return jsonutils.loads(response.text)
        return response.text

    def _check_artifact_post(self, url, data, status=201,
                             headers=None):
        if headers is None:
            headers = {'Content-Type': 'application/json'}

        return self._check_artifact_method("post", url, data, status=status,
                                           headers=headers)

    def _check_artifact_get(self, url, status=200):
        return self._check_artifact_method("get", url, status=status)

    def _check_artifact_delete(self, url, status=204):
        response = requests.delete(self._url(url), headers=self._headers())
        self.assertEqual(status, response.status_code)
        return response.text

    def _check_artifact_patch(self, url, data, status=200,
                              headers={'Content-Type': 'application/json'}):
        return self._check_artifact_method("patch", url, data, status=status,
                                           headers=headers)

    def _check_artifact_put(self, url, data, status=200,
                            headers={'Content-Type': 'application/json'}):
        return self._check_artifact_method("put", url, data, status=status,
                                           headers=headers)

    def test_list_any_artifacts(self):
        """Returns information about all draft artifacts with given endpoint"""
        self._create_artifact('noprop')
        artifacts = self._check_artifact_get('/noprop/drafts')["artifacts"]
        self.assertEqual(1, len(artifacts))

    def test_list_last_version(self):
        """/artifacts/endpoint == /artifacts/endpoint/all-versions"""
        self._create_artifact('noprop')
        artifacts = self._check_artifact_get('/noprop/drafts')["artifacts"]
        self.assertEqual(1, len(artifacts))
        # the same result can be achieved if asked for artifact with
        # type_version=last version
        artifacts_precise = self._check_artifact_get(
            '/noprop/v1.0/drafts')["artifacts"]
        self.assertEqual(artifacts, artifacts_precise)

    def test_list_artifacts_by_state(self):
        """Returns last version of artifacts with given state"""
        self._create_artifact('noprop')
        creating_state = self._check_artifact_get(
            '/noprop/drafts')["artifacts"]
        self.assertEqual(1, len(creating_state))
        # no active [/type_name/active == /type_name]
        active_state = self._check_artifact_get('/noprop')["artifacts"]
        self.assertEqual(0, len(active_state))

    def test_list_artifacts_with_version(self):
        """Supplying precise artifact version does not break anything"""
        self._create_artifact('noprop')
        list_creating = self._check_artifact_get(
            '/noprop/v1.0/drafts')["artifacts"]
        self.assertEqual(1, len(list_creating))
        bad_version = self._check_artifact_get('/noprop/v1.0bad',
                                               status=400)
        self.assertIn("Invalid version string: u'1.0bad'", bad_version)

    def test_list_artifacts_with_pagination(self):
        """List artifacts with pagination"""
        # create artifacts
        art1 = {'name': 'artifact-1',
                'version': '12'}
        art2 = {'name': 'artifact-2',
                'version': '12'}
        self._create_artifact('noprop', data=art1)
        self._create_artifact('noprop', data=art2)
        # sorting is desc by default
        first_page = self._check_artifact_get(
            '/noprop/drafts?limit=1&sort=name')
        # check the first artifacts has returned correctly
        self.assertEqual(1, len(first_page["artifacts"]))
        self.assertEqual("artifact-2", first_page["artifacts"][0]["name"])
        self.assertIn("next", first_page)
        # check the second page
        second_page_url = first_page["next"].split("artifacts", 1)[1]
        second_page = self._check_artifact_get(second_page_url)
        self.assertIn("next", second_page)
        self.assertEqual(1, len(second_page["artifacts"]))
        self.assertEqual("artifact-1", second_page["artifacts"][0]["name"])
        # check that the latest item is empty
        last_page_url = second_page["next"].split("artifacts", 1)[1]
        last_page = self._check_artifact_get(last_page_url)
        self.assertEqual(0, len(last_page["artifacts"]))
        self.assertNotIn("next", last_page)

    def test_get_artifact_by_id_any_version(self):
        data = self._create_artifact('noprop')
        artifact_id = data['id']
        artifacts = self._check_artifact_get(
            '/noprop/%s' % artifact_id)
        self.assertEqual(artifact_id, artifacts['id'])

    def test_list_artifact_no_such_version(self):
        """Version filtering should be applied for existing plugins.

        An attempt to retrieve an artifact out of existing plugin but with
        a wrong version should result in
        400 BadRequest 'No such plugin has been loaded'
        """
        msg = self._check_artifact_get('/noprop/v0.0.9', 400)
        self.assertIn("No plugin for 'noprop v 0.0.9' has been loaded",
                      msg)

    def test_get_artifact_by_id(self):
        data = self._create_artifact('noprop')
        artifact_id = data['id']
        artifacts = self._check_artifact_get(
            '/noprop/%s' % artifact_id)
        self.assertEqual(artifact_id, artifacts['id'])
        # the same result can be achieved if asked for artifact with
        # type_version=last version
        artifacts_precise = self._check_artifact_get(
            '/noprop/v1.0/%s' % artifact_id)
        self.assertEqual(artifacts, artifacts_precise)

    def test_get_artifact_basic_show_level(self):
        no_prop_art = self._create_artifact('noprop')
        art = self._create_artifact(
            'withprops',
            data={"name": "name", "version": "42",
                  "depends_on": no_prop_art['id']})
        self.assertEqual(no_prop_art['id'], art['depends_on']['id'])
        self.assertEqual(no_prop_art['name'], art['depends_on']['name'])

        artifact_id = art['id']
        artifact = self._check_artifact_get(
            '/withprops/%s?show_level=basic' % artifact_id)
        self.assertEqual(artifact_id, artifact['id'])
        self.assertIsNone(artifact['depends_on'])

    def test_get_artifact_none_show_level(self):
        """Create an artifact (with two deployer-defined properties)"""
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'tags': ['gagaga', 'sesese'],
                         'prop1': 'Arthur Dent',
                         'prop2': 42}
        art = self._check_artifact_post('/withprops/v1.0/drafts',
                                        artifact_data)
        expected_artifact = {
            'state': 'creating',
            'name': 'artifact-1',
            'version': '12.0.0',
            'tags': ['gagaga', 'sesese'],
            'visibility': 'private',
            'type_name': 'WithProps',
            'type_version': '1.0',
            'prop1': 'Arthur Dent',
            'prop2': 42
        }
        for key, value in expected_artifact.items():
            self.assertEqual(art[key], value, key)

        artifact_id = art['id']
        artifact = self._check_artifact_get(
            '/withprops/%s?show_level=none' % artifact_id)
        self.assertEqual(artifact_id, artifact['id'])
        self.assertIsNone(artifact['prop1'])
        self.assertIsNone(artifact['prop2'])

    def test_get_artifact_invalid_show_level(self):
        no_prop_art = self._create_artifact('noprop')
        art = self._create_artifact(
            'withprops',
            data={"name": "name", "version": "42",
                  "depends_on": no_prop_art['id']})
        self.assertEqual(no_prop_art['id'], art['depends_on']['id'])
        self.assertEqual(no_prop_art['name'], art['depends_on']['name'])

        artifact_id = art['id']
        # 'hui' is invalid show level
        self._check_artifact_get(
            '/noprop/%s?show_level=yoba' % artifact_id, status=400)

    def test_get_artifact_no_such_id(self):
        msg = self._check_artifact_get(
            '/noprop/%s' % str(uuid.uuid4()), status=404)
        self.assertIn('No artifact found with ID', msg)

    def test_get_artifact_present_id_wrong_type(self):
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'prop1': '12',
                         'prop2': 12}
        art1 = self._create_artifact('withprops', data=artifact_data)
        art2 = self._create_artifact('noprop')
        # ok id and type_name but bad type_version should result in 404
        self._check_artifact_get('/noprop/v0.5/%s' % str(art2['id']),
                                 status=404)
        # try to access art2 by supplying art1.type and art2.id
        self._check_artifact_get('/withprops/%s' % str(art2['id']),
                                 status=404)
        self._check_artifact_get('/noprop/%s' % str(art1['id']), status=404)

    def test_delete_artifact(self):
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'prop1': '12',
                         'prop2': 12}
        art1 = self._create_artifact('withprops', data=artifact_data)
        self._check_artifact_delete('/withprops/v1.0/%s' % art1['id'])
        art1_deleted = self._check_artifact_get('/withprops/%s' % art1['id'],
                                                status=404)
        self.assertIn('No artifact found with ID', art1_deleted)

    def test_delete_artifact_no_such_id(self):
        self._check_artifact_delete('/noprop/v1/%s' % str(uuid.uuid4()),
                                    status=404)

    @unittest.skip("Test is unstable")
    def test_delete_artifact_with_dependency(self):
        # make sure that artifact can't be deleted if it has some dependencies
        # still not deleted
        art = self._create_artifact('withprops')
        no_prop_art = self._create_artifact('noprop')
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'value': no_prop_art['id'],
                   'op': 'replace',
                   'path': '/depends_on'},
                  {'value': [no_prop_art['id']],
                   'op': 'add',
                   'path': '/depends_on_list'}])
        self.assertEqual(no_prop_art['id'], art_updated['depends_on']['id'])
        self.assertEqual(1, len(art_updated['depends_on_list']))
        # try to delete an artifact prior to its dependency
        res = self._check_artifact_delete('/withprops/v1/%s' % art['id'],
                                          status=400)
        self.assertIn(
            "Dependency property 'depends_on' has to be deleted first", res)
        # delete a dependency
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'op': 'remove', 'path': '/depends_on'}])
        # try to delete prior to deleting artifact_list dependencies
        res = self._check_artifact_delete('/withprops/v1/%s' % art['id'],
                                          status=400)
        self.assertIn(
            "Dependency property 'depends_on_list' has to be deleted first",
            res)
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'op': 'remove', 'path': '/depends_on_list'}])
        # delete dependency list
        self._check_artifact_delete('/withprops/v1/%s' % art['id'])

    def test_delete_artifact_with_blob(self):
        # Upload some data to an artifact
        art = self._create_artifact('withblob')
        headers = self._headers({'Content-Type': 'application/octet-stream'})
        self._check_artifact_post('/withblob/v1/%s/blob1' % art['id'],
                                  headers=headers,
                                  data='ZZZZZ', status=200)
        self._check_artifact_delete('/withblob/v1/%s' % art['id'])

    def test_update_nonexistent_property_by_replace_op(self):
        art = self._create_artifact('withprops', data={'name': 'some art',
                                                       'version': '4.2'})
        data = [{'op': 'replace', 'value': 'some value',
                 'path': '/nonexistent_property'}]
        result = self._check_artifact_patch('/withprops/v1/%s' %
                                            art['id'],
                                            data=data,
                                            status=400)
        self.assertIn('400 Bad Request', result)
        self.assertIn('Artifact has no property nonexistent_property', result)

    def test_update_nonexistent_property_by_remove_op(self):
        art = self._create_artifact('withprops', data={'name': 'some art',
                                                       'version': '4.2'})
        data = [{'op': 'replace', 'value': 'some value',
                 'path': '/nonexistent_property'}]
        result = self._check_artifact_patch('/withprops/v1/%s' %
                                            art['id'],
                                            data=data,
                                            status=400)
        self.assertIn('400 Bad Request', result)
        self.assertIn('Artifact has no property nonexistent_property', result)

    def test_update_nonexistent_property_by_add_op(self):
        art = self._create_artifact('withprops', data={'name': 'some art',
                                                       'version': '4.2'})
        data = [{'op': 'replace', 'value': 'some value',
                 'path': '/nonexistent_property'}]
        result = self._check_artifact_patch('/withprops/v1/%s' %
                                            art['id'],
                                            data=data,
                                            status=400)
        self.assertIn('400 Bad Request', result)
        self.assertIn('Artifact has no property nonexistent_property', result)

    def test_update_array_property_by_replace_op(self):
        art = self._create_artifact('withprops', data={'name': 'some art',
                                                       'version': '4.2'})
        self.assertEqual('some art', art['name'])
        data = [{'op': 'replace', 'value': [1, 2, 3], 'path': '/prop_list'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s' %
                                                 art['id'],
                                                 data=data)
        self.assertEqual([1, 2, 3], art_updated['prop_list'])
        # now try to change first element of the list
        data_change_first = [{'op': 'replace', 'value': 42,
                              'path': '/prop_list/1'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s' %
                                                 art['id'],
                                                 data=data_change_first)
        self.assertEqual([1, 42, 3], art_updated['prop_list'])
        # replace last element
        data_change_last = [{'op': 'replace', 'value': 24,
                             'path': '/prop_list/-'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s' %
                                                 art['id'],
                                                 data=data_change_last)
        self.assertEqual([1, 42, 24], art_updated['prop_list'])

    def test_update_dict_property_by_replace_op(self):
        art = self._create_artifact(
            'withprops',
            data={'name': 'some art',
                  'version': '4.2',
                  'dict_prop': {'foo': "Fenchurch", 'bar_list': [42, 42]}})
        self.assertEqual({'foo': "Fenchurch", 'bar_list': [42, 42]},
                         art['dict_prop'])
        data = [{'op': 'replace', 'value': 24,
                 'path': '/dict_prop/bar_list/0'},
                {'op': 'replace', 'value': 'cello lesson',
                 'path': '/dict_prop/foo'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual({'foo': 'cello lesson', 'bar_list': [24, 42]},
                         art_updated['dict_prop'])

    def test_update_empty_dict_property_by_replace_op(self):
        art = self._create_artifact('withprops')
        self.assertIsNone(art['dict_prop'])
        data = [{'op': 'replace', 'value': "don't panic",
                 'path': '/dict_prop/foo'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data, status=400)
        self.assertIn("The provided path 'dict_prop/foo' is invalid",
                      art_updated)

    def test_update_empty_dict_property_by_remove_op(self):
        art = self._create_artifact('withprops')
        self.assertIsNone(art['dict_prop'])
        data = [{'op': 'remove', 'path': '/dict_prop/bar_list'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data, status=400)
        self.assertIn("The provided path 'dict_prop/bar_list' is invalid",
                      art_updated)

    def test_update_dict_property_by_remove_op(self):
        art = self._create_artifact(
            'withprops',
            data={'name': 'some art', 'version': '4.2',
                  'dict_prop': {'foo': "Fenchurch", 'bar_list': [42, 42]}})
        self.assertEqual({'foo': 'Fenchurch', 'bar_list': [42, 42]},
                         art['dict_prop'])
        data = [{'op': 'remove', 'path': '/dict_prop/foo'},
                {'op': 'remove', 'path': '/dict_prop/bar_list/1'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual({'bar_list': [42]}, art_updated['dict_prop'])
        # now delete the whole dict
        data = [{'op': 'remove', 'path': '/dict_prop'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertIsNone(art_updated['dict_prop'])

    @unittest.skip("Skipping due to a know bug")
    def test_update_dict_property_change_values(self):
        art = self._create_artifact(
            'withprops', data={'name': 'some art', 'version': '4.2',
                               'dict_prop_strval':
                               {'foo': 'Fenchurch', 'bar': 'no value'}})
        self.assertEqual({'foo': 'Fenchurch', 'bar': 'no value'},
                         art['dict_prop_strval'])
        new_data = [{'op': 'replace', 'path': '/dict_prop_strval',
                     'value': {'new-foo': 'Arthur Dent'}}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=new_data)
        self.assertEqual({'new-foo': 'Arthur Dent'},
                         art_updated['dict_prop_strval'])

    def test_update_array_property_by_remove_op(self):
        art = self._create_artifact(
            'withprops', data={'name': 'some art',
                               'version': '4.2',
                               'prop_list': [1, 2, 3]})
        self.assertEqual([1, 2, 3], art['prop_list'])
        data = [{'op': 'remove', 'path': '/prop_list/0'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual([2, 3], art_updated['prop_list'])
        # remove last element
        data = [{'op': 'remove', 'path': '/prop_list/-'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual([2], art_updated['prop_list'])
        # now delete the whole array
        data = [{'op': 'remove', 'path': '/prop_list'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertIsNone(art_updated['prop_list'])

    def test_update_array_property_by_add_op(self):
        art = self._create_artifact(
            'withprops', data={'name': 'some art',
                               'version': '4.2'})
        self.assertIsNone(art['prop_list'])
        data = [{'op': 'add', 'path': '/prop_list', 'value': [2, 12, 0, 6]}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'], data=data)
        self.assertEqual([2, 12, 0, 6], art_updated['prop_list'])
        data = [{'op': 'add', 'path': '/prop_list/2', 'value': 85}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'], data=data)
        self.assertEqual([2, 12, 85, 0, 6], art_updated['prop_list'])
        # add where path='/array/-' means append to the end
        data = [{'op': 'add', 'path': '/prop_list/-', 'value': 7}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'], data=data)
        self.assertEqual([2, 12, 85, 0, 6, 7], art_updated['prop_list'])
        # an attempt to add an element to nonexistent position should
        # result in 400
        self.assertEqual(6, len(art_updated['prop_list']))
        bad_index_data = [{'op': 'add', 'path': '/prop_list/11',
                           'value': 42}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=bad_index_data,
                                                 status=400)
        self.assertIn("The provided path 'prop_list/11' is invalid",
                      art_updated)

    def test_update_dict_property_by_add_op(self):
        art = self._create_artifact("withprops")
        self.assertIsNone(art['dict_prop'])
        data = [{'op': 'add', 'path': '/dict_prop/foo', 'value': "some value"}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual({"foo": "some value"}, art_updated['dict_prop'])

    def test_update_empty_array_property_by_add_op(self):
        """Test jsonpatch add.

        According to RFC 6902:
        * if the array is empty, '/array/0' is a valid path
        """
        create_data = {'name': 'new artifact',
                       'version': '4.2'}
        art = self._create_artifact('withprops', data=create_data)
        self.assertIsNone(art['prop_list'])
        data = [{'op': 'add', 'path': '/prop_list/0', 'value': 3}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual([3], art_updated['prop_list'])

    def test_update_tuple_property_by_index(self):
        art = self._create_artifact(
            'withprops', data={'name': 'some art',
                               'version': '4.2',
                               'tuple_prop': [1, False]})
        self.assertEqual([1, False], art['tuple_prop'])
        data = [{'op': 'replace', 'value': True,
                 'path': '/tuple_prop/1'},
                {'op': 'replace', 'value': 2,
                 'path': '/tuple_prop/0'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertEqual([2, True], art_updated['tuple_prop'])

    def test_update_artifact(self):
        art = self._create_artifact('noprop')
        self.assertEqual('artifact-1', art['name'])
        art_updated = self._check_artifact_patch(
            '/noprop/v1/%s' % art['id'],
            data=[{'op': 'replace', 'value': '0.0.9', 'path': '/version'}])
        self.assertEqual('0.0.9', art_updated['version'])

    def test_update_artifact_properties(self):
        art = self._create_artifact('withprops')
        for prop in ['prop1', 'prop2']:
            self.assertIsNone(art[prop])
            data = [{'op': 'replace', 'value': 'some value',
                     'path': '/prop1'}]
            art_updated = self._check_artifact_patch(
                '/withprops/v1/%s' % art['id'], data=data)
            self.assertEqual('some value', art_updated['prop1'])

    def test_update_remove_non_existent_artifact_properties(self):
        art = self._create_artifact('withprops')
        for prop in ['prop1', 'prop2']:
            self.assertIsNone(art[prop])
            data = [{'op': 'remove', 'value': 'some value',
                     'path': '/non-existent-path/and-another'}]
            art_updated = self._check_artifact_patch(
                '/withprops/v1/%s' % art['id'], data=data, status=400)
            self.assertIn('Artifact has no property', art_updated)

    def test_update_replace_non_existent_artifact_properties(self):
        art = self._create_artifact('withprops')
        for prop in ['prop1', 'prop2']:
            self.assertIsNone(art[prop])
            data = [{'op': 'replace', 'value': 'some value',
                     'path': '/non-existent-path/and-another'}]
            art_updated = self._check_artifact_patch(
                '/withprops/v1/%s' % art['id'], data=data, status=400)
            self.assertIn('Artifact has no property', art_updated)

    def test_update_artifact_remove_property(self):
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'tags': ['gagaga', 'sesese'],
                         'prop1': 'Arthur Dent',
                         'prop2': 42}
        art = self._create_artifact('withprops', data=artifact_data)
        data = [{'op': 'remove', 'path': '/prop1'}]
        art_updated = self._check_artifact_patch('/withprops/v1/%s'
                                                 % art['id'],
                                                 data=data)
        self.assertIsNone(art_updated['prop1'])
        self.assertEqual(42, art_updated['prop2'])

    def test_update_wrong_property_type(self):
        art = self._create_artifact('withprops')
        for prop in ['prop2', 'prop2']:
            self.assertIsNone(art[prop])
        data = [{'op': 'replace', 'value': 123, 'path': '/prop1'}]
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'], data=data, status=400)
        self.assertIn("Property 'prop1' may not have value '123'", art_updated)

    def test_update_multiple_properties(self):
        with_prop_art = self._create_artifact('withprops')
        data = [{'op': 'replace',
                 'path': '/prop1',
                 'value': 'some value'},
                {'op': 'replace',
                 'path': '/prop2',
                 'value': 42}]
        updated = self._check_artifact_patch(
            '/withprops/v1/%s' % with_prop_art['id'], data=data)
        self.assertEqual('some value', updated['prop1'])
        self.assertEqual(42, updated['prop2'])

    def test_create_artifact_with_dependency(self):
        no_prop_art = self._create_artifact('noprop')
        art = self._create_artifact(
            'withprops',
            data={"name": "name", "version": "42",
                  "depends_on": no_prop_art['id']})
        self.assertEqual(no_prop_art['id'], art['depends_on']['id'])
        self.assertEqual(no_prop_art['name'], art['depends_on']['name'])

    def test_create_artifact_dependency_list(self):
        no_prop_art1 = self._create_artifact('noprop')
        no_prop_art2 = self._create_artifact('noprop')
        art = self._create_artifact(
            'withprops',
            data={"name": "name", "version": "42",
                  "depends_on_list": [no_prop_art1['id'], no_prop_art2['id']]})
        self.assertEqual(2, len(art['depends_on_list']))
        self.assertEqual([no_prop_art1['id'], no_prop_art2['id']],
                         map(lambda x: x['id'], art['depends_on_list']))

    def test_create_dependency_list_same_id(self):
        no_prop_art = self._create_artifact('noprop')
        res = self._create_artifact(
            'withprops',
            data={"name": "name", "version": "42",
                  "depends_on_list": [no_prop_art['id'],
                                      no_prop_art['id']]}, status=400)
        self.assertIn("Items have to be unique", res)

    def test_create_artifact_bad_dependency_format(self):
        """Invalid dependencies creation.

        Dependencies should be passed:
        * as a list of ids if param is an ArtifactReferenceList
        * as an id if param is an ArtifactReference
        """
        no_prop_art = self._create_artifact('noprop')
        art = self._check_artifact_post(
            '/withprops/v1/drafts',
            {"name": "name", "version": "42",
             "depends_on": [no_prop_art['id']]}, status=400)
        self.assertIn('Not a valid value type', art)
        art = self._check_artifact_post(
            '/withprops/v1.0/drafts',
            {"name": "name", "version": "42",
             "depends_on_list": no_prop_art['id']}, status=400)
        self.assertIn('object is not iterable', art)

    def test_update_dependency(self):
        no_prop_art = self._create_artifact('noprop')
        no_prop_art1 = self._create_artifact('noprop')
        with_prop_art = self._create_artifact('withprops')
        data = [{'op': 'replace',
                 'path': '/depends_on',
                 'value': no_prop_art['id']}]
        updated = self._check_artifact_patch(
            '/withprops/v1/%s' % with_prop_art['id'], data=data)
        self.assertEqual(no_prop_art['id'], updated['depends_on']['id'])
        self.assertEqual(no_prop_art['name'], updated['depends_on']['name'])
        data = [{'op': 'replace',
                 'path': '/depends_on',
                 'value': no_prop_art1['id']}]
        # update again and make sure it changes
        updated = self._check_artifact_patch(
            '/withprops/v1/%s' % with_prop_art['id'], data=data)
        self.assertEqual(no_prop_art1['id'], updated['depends_on']['id'])
        self.assertEqual(no_prop_art1['name'], updated['depends_on']['name'])

    def test_update_dependency_circular_reference(self):
        with_prop_art = self._create_artifact('withprops')
        data = [{'op': 'replace',
                 'path': '/depends_on',
                 'value': [with_prop_art['id']]}]
        not_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % with_prop_art['id'], data=data, status=400)
        self.assertIn('Artifact with a circular dependency can not be created',
                      not_updated)

    def test_publish_artifact(self):
        art = self._create_artifact('withprops')
        # now create dependency
        no_prop_art = self._create_artifact('noprop')
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'value': no_prop_art['id'],
                   'op': 'replace',
                   'path': '/depends_on'}])
        self.assertNotEqual(0, len(art_updated['depends_on']))
        # artifact can't be published if any dependency is in non-active state
        res = self._check_artifact_post(
            '/withprops/v1/%s/publish' % art['id'], {}, status=400)
        self.assertIn("Not all dependencies are in 'active' state", res)
        # after you publish the dependency -> artifact can be published
        dep_published = self._check_artifact_post(
            '/noprop/v1/%s/publish' % no_prop_art['id'], {}, status=200)
        self.assertEqual('active', dep_published['state'])
        art_published = self._check_artifact_post(
            '/withprops/v1.0/%s/publish' % art['id'], {}, status=200)
        self.assertEqual('active', art_published['state'])

    def test_no_mutable_change_in_published_state(self):
        art = self._create_artifact('withprops')
        no_prop_art = self._create_artifact('noprop')
        no_prop_other = self._create_artifact('noprop')
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'value': no_prop_art['id'],
                   'op': 'replace',
                   'path': '/depends_on'}])
        self.assertEqual(no_prop_art['id'], art_updated['depends_on']['id'])
        # now change dependency to some other artifact
        art_updated = self._check_artifact_patch(
            '/withprops/v1/%s' % art['id'],
            data=[{'value': no_prop_other['id'],
                   'op': 'replace',
                   'path': '/depends_on'}])
        self.assertEqual(no_prop_other['id'], art_updated['depends_on']['id'])
        # publish dependency
        dep_published = self._check_artifact_post(
            '/noprop/v1/%s/publish' % no_prop_other['id'], {}, status=200)
        self.assertEqual('active', dep_published['state'])
        # publish artifact
        art_published = self._check_artifact_post(
            '/withprops/v1.0/%s/publish' % art['id'], {}, status=200)
        self.assertEqual('active', art_published['state'])
        # try to change dependency, should fail as already published
        res = self._check_artifact_patch(
            '/withprops/v1/%s' % art_published['id'],
            data=[{'op': 'remove', 'path': '/depends_on'}], status=400)
        self.assertIn('Attempt to set value of immutable property', res)

    def test_create_artifact_empty_body(self):
        self._check_artifact_post('/noprop/v1.0/drafts', {}, 400)

    def test_create_artifact_insufficient_arguments(self):
        self._check_artifact_post('/noprop/v1.0/drafts',
                                  {'name': 'some name, no version'},
                                  status=400)

    def test_create_artifact_no_such_version(self):
        """Creation impossible without specifying a correct version.

        An attempt to create an artifact out of existing plugin but with
        a wrong version should result in
        400 BadRequest 'No such plugin has been loaded'
        """
        # make sure there is no such artifact noprop
        self._check_artifact_get('/noprop/v0.0.9', 400)
        artifact_data = {'name': 'artifact-1',
                         'version': '12'}
        msg = self._check_artifact_post('/noprop/v0.0.9/drafts',
                                        artifact_data,
                                        status=400)
        self.assertIn("No plugin for 'noprop v 0.0.9' has been loaded",
                      msg)

    def test_create_artifact_no_type_version_specified(self):
        """Creation impossible without specifying a version.

        It should not be possible to create an artifact out of existing plugin
        without specifying any version
        """
        artifact_data = {'name': 'artifact-1',
                         'version': '12'}
        self._check_artifact_post('/noprop/drafts', artifact_data, 404)

    def test_create_artifact_no_properties(self):
        """Create an artifact with minimum parameters"""
        artifact_data = {'name': 'artifact-1',
                         'version': '12'}
        artifact = self._check_artifact_post('/withprops/v1.0/drafts',
                                             artifact_data)
        # verify that all fields have the values expected
        expected_artifact = {
            'state': 'creating',
            'name': 'artifact-1',
            'version': '12.0.0',
            'tags': [],
            'visibility': 'private',
            'type_name': 'WithProps',
            'type_version': '1.0',
            'prop1': None,
            'prop2': None
        }
        for key, value in expected_artifact.items():
            self.assertEqual(artifact[key], value, key)

    def test_create_artifact_with_properties(self):
        """Create an artifact (with two deployer-defined properties)"""
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'tags': ['gagaga', 'sesese'],
                         'prop1': 'Arthur Dent',
                         'prop2': 42}
        artifact = self._check_artifact_post('/withprops/v1.0/drafts',
                                             artifact_data)
        expected_artifact = {
            'state': 'creating',
            'name': 'artifact-1',
            'version': '12.0.0',
            'tags': ['gagaga', 'sesese'],
            'visibility': 'private',
            'type_name': 'WithProps',
            'type_version': '1.0',
            'prop1': 'Arthur Dent',
            'prop2': 42
        }
        for key, value in expected_artifact.items():
            self.assertEqual(artifact[key], value, key)

    def test_create_artifact_not_all_properties(self):
        """Create artifact with minimal properties.

        Checks that it is possible to create an artifact by passing all
        required properties but omitting some not required
        """
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'visibility': 'private',
                         'tags': ['gagaga', 'sesese'],
                         'prop1': 'i am ok'}
        artifact = self._check_artifact_post('/withprops/v1.0/drafts',
                                             artifact_data)
        expected_artifact = {
            'state': 'creating',
            'name': 'artifact-1',
            'version': '12.0.0',
            'tags': ['gagaga', 'sesese'],
            'visibility': 'private',
            'type_name': 'WithProps',
            'type_version': '1.0',
            'prop1': 'i am ok',
            'prop2': None}
        for key, value in expected_artifact.items():
            self.assertEqual(artifact[key], value, key)
        # now check creation with no properties specified
        for prop in ['prop1', 'prop2']:
            artifact_data.pop(prop, '')
        artifact = self._check_artifact_post('/withprops/v1.0/drafts',
                                             artifact_data)
        for prop in ['prop1', 'prop2']:
            self.assertIsNone(artifact[prop])

    def test_create_artifact_invalid_properties(self):
        """Any attempt to pass invalid properties should result in 400"""
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'prop1': 1}
        res = self._check_artifact_post('/withprops/v1.0/drafts',
                                        artifact_data,
                                        status=400)
        self.assertIn("Property 'prop1' may not have value '1'", res)
        artifact_data.pop('prop1')
        artifact_data['nosuchprop'] = "Random"
        res = self._check_artifact_post('/withprops/v1.0/drafts',
                                        artifact_data,
                                        status=400)
        self.assertIn("Artifact has no property nosuchprop", res)

    def test_create_public_artifact(self):
        """Create an artifact with visibility set to public"""
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'visibility': 'public'}
        artifact = self._check_artifact_post('/withprops/v1.0/drafts',
                                             artifact_data)
        # verify that all fields have the values expected
        expected_artifact = {
            'state': 'creating',
            'name': 'artifact-1',
            'version': '12.0.0',
            'tags': [],
            'visibility': 'public',
            'type_name': 'WithProps',
            'type_version': '1.0',
            'prop1': None,
            'prop2': None
        }
        for key, value in expected_artifact.items():
            self.assertEqual(artifact[key], value, key)

    def test_upload_file(self):
        # Upload some data to an artifact
        art = self._create_artifact('withblob')
        headers = self._headers({'Content-Type': 'application/octet-stream'})
        self._check_artifact_post('/withblob/v1/%s/blob1' % art['id'],
                                  headers=headers,
                                  data='ZZZZZ', status=200)

    def test_upload_file_with_invalid_content_type(self):
        art = self._create_artifact('withblob')
        data = {'data': 'jjjjjj'}
        res = self._check_artifact_post('/withblob/v1/%s/blob1' % art['id'],
                                        data=data, status=400)
        self.assertIn('Invalid Content-Type for work with blob1', res)

        res = self._check_artifact_post('/withblob/v1/%s/blob_list'
                                        % art['id'],
                                        data=data, status=400)
        self.assertIn('Invalid Content-Type for work with blob_list', res)

    def test_upload_list_files(self):
        art = self._create_artifact('withblob')
        headers = self._headers({'Content-Type': 'application/octet-stream'})
        self._check_artifact_post('/withblob/v1/%s/blob_list' % art['id'],
                                  headers=headers,
                                  data='ZZZZZ', status=200)
        self._check_artifact_post('/withblob/v1/%s/blob_list' % art['id'],
                                  headers=headers,
                                  data='YYYYY', status=200)

    def test_download_file(self):
        # Download some data from an artifact
        art = self._create_artifact('withblob')
        artifact_id = art['id']
        headers = self._headers({'Content-Type': 'application/octet-stream'})
        self._check_artifact_post('/withblob/v1/%s/blob1' % art['id'],
                                  headers=headers,
                                  data='ZZZZZ', status=200)

        art = self._check_artifact_get('/withblob/%s' % artifact_id)
        self.assertEqual(artifact_id, art['id'])
        self.assertIn('download_link', art['blob1'])

        data = self._check_artifact_get(
            '/withblob/%s/blob1/download' % art['id'])
        self.assertEqual('ZZZZZ', data)

    def test_file_w_unknown_size(self):
        # Upload and download data provided by an iterator, thus without
        # knowing the length in advance
        art = self._create_artifact('withblob')
        artifact_id = art['id']

        def iterate_string(val):
            for char in val:
                yield char

        headers = self._headers({'Content-Type': 'application/octet-stream'})
        self._check_artifact_post('/withblob/v1/%s/blob1' % art['id'],
                                  headers=headers,
                                  data=iterate_string('ZZZZZ'), status=200)

        art = self._check_artifact_get('/withblob/%s' % artifact_id)
        self.assertEqual(artifact_id, art['id'])
        self.assertIn('download_link', art['blob1'])

        data = self._check_artifact_get(
            '/withblob/%s/blob1/download' % art['id'])
        self.assertEqual('ZZZZZ', data)

    def test_limit(self):
        artifact_data = {'name': 'artifact-1',
                         'version': '12'}
        self._check_artifact_post('/withprops/v1/drafts',
                                  artifact_data)
        artifact_data = {'name': 'artifact-1',
                         'version': '13'}
        self._check_artifact_post('/withprops/v1/drafts',
                                  artifact_data)
        result = self._check_artifact_get('/withprops/v1/drafts')
        self.assertEqual(2, len(result["artifacts"]))
        result = self._check_artifact_get('/withprops/v1/drafts?limit=1')
        self.assertEqual(1, len(result["artifacts"]))

    def _check_sorting_order(self, expected, actual):
        for e, a in zip(expected, actual):
            self.assertEqual(e['name'], a['name'])
            self.assertEqual(e['version'], a['version'])
            self.assertEqual(e['prop1'], a['prop1'])

    def test_sort(self):
        artifact_data = {'name': 'artifact-1',
                         'version': '12',
                         'prop1': 'lala'}
        art1 = self._check_artifact_post('/withprops/v1.0/drafts',
                                         artifact_data)
        artifact_data = {'name': 'artifact-2',
                         'version': '13',
                         'prop1': 'lala'}
        art2 = self._check_artifact_post('/withprops/v1.0/drafts',
                                         artifact_data)
        artifact_data = {'name': 'artifact-3',
                         'version': '13',
                         'prop1': 'tutu'}
        art3 = self._check_artifact_post('/withprops/v1.0/drafts',
                                         artifact_data)
        artifact_data = {'name': 'artifact-4',
                         'version': '13',
                         'prop1': 'hyhy'}
        art4 = self._check_artifact_post('/withprops/v1.0/drafts',
                                         artifact_data)
        artifact_data = {'name': 'artifact-5',
                         'version': '13',
                         'prop1': 'bebe'}
        art5 = self._check_artifact_post('/withprops/v1.0/drafts',
                                         artifact_data)

        result = self._check_artifact_get(
            '/withprops/v1.0/drafts?sort=name')["artifacts"]
        self.assertEqual(5, len(result))

        # default direction is 'desc'
        expected = [art5, art4, art3, art2, art1]
        self._check_sorting_order(expected, result)

        result = self._check_artifact_get(
            '/withprops/v1.0/drafts?sort=name:asc')["artifacts"]
        self.assertEqual(5, len(result))

        expected = [art1, art2, art3, art4, art5]
        self._check_sorting_order(expected, result)

        result = self._check_artifact_get(
            '/withprops/v1.0/drafts?sort=version:asc,prop1')["artifacts"]
        self.assertEqual(5, len(result))

        expected = [art1, art3, art2, art4, art5]
        self._check_sorting_order(expected, result)

    def test_update_property(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        # update single integer property via PUT
        upd = self._check_artifact_put('/withprops/v1.0/%s/prop2' % art['id'],
                                       data={'data': 15})
        self.assertEqual(15, upd['prop2'])
        # create list property via PUT
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/tuple_prop' % art['id'],
            data={'data': [42, True]})
        self.assertEqual([42, True], upd['tuple_prop'])
        # change list property via PUT
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/tuple_prop/0' % art['id'], data={'data': 24})
        self.assertEqual([24, True], upd['tuple_prop'])
        # append to list property via POST
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/prop_list' % art['id'], data={'data': [11]},
            status=200)
        self.assertEqual([11], upd['prop_list'])
        # append to list property via POST
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/prop_list/-' % art['id'],
            status=200, data={'data': 10})
        self.assertEqual([11, 10], upd['prop_list'])

    def test_bad_update_property(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        # try to update nonexistent property
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/nosuchprop' % art['id'],
            data={'data': 'wont be set'}, status=400)
        self.assertIn('Artifact has no property nosuchprop', upd)
        # try to pass wrong property value
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/tuple_prop' % art['id'],
            data={'data': ['should be an int', False]}, status=400)
        self.assertIn("Property 'tuple_prop[0]' may not have value", upd)
        # try to pass bad body (not a valid json)
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/tuple_prop' % art['id'], data="not a json",
            status=400)
        self.assertIn("Invalid json body", upd)
        # try to pass json body invalid under schema
        upd = self._check_artifact_put(
            '/withprops/v1.0/%s/tuple_prop' % art['id'],
            data={"bad": "schema"}, status=400)
        self.assertIn("Invalid json body", upd)

    def test_update_different_depths_levels(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/dict_prop' % art['id'],
            data={'data': {'foo': 'some value'}}, status=200)
        self.assertEqual({'foo': 'some value'}, upd['dict_prop'])
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/dict_prop/bar_list' % art['id'],
            data={'data': [5]}, status=200)
        self.assertEqual({'foo': 'some value', 'bar_list': [5]},
                         upd['dict_prop'])
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/dict_prop/bar_list/0' % art['id'],
            data={'data': 15}, status=200)
        self.assertEqual({'foo': 'some value', 'bar_list': [5, 15]},
                         upd['dict_prop'])
        # try to attempt dict_property by nonexistent path
        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/dict_prop/bar_list/nosuchkey' % art['id'],
            data={'data': 15}, status=400)

    def test_artifact_inaccessible_by_different_user(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        self._set_user('user2')
        self._check_artifact_get('/withprops/%s' % art['id'], 404)

    def test_artifact_accessible_by_admin(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        self._set_user('admin')
        self._check_artifact_get('/withprops/%s' % art['id'], 200)

    def test_public_artifact_accessible_by_different_user(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        self._check_artifact_patch(
            '/withprops/v1.0/%s' % art['id'],
            data=[{'op': 'replace', 'value': 'public', 'path': '/visibility'}])
        self._set_user('user2')
        self._check_artifact_get('/withprops/%s' % art['id'], 200)

    def test_public_artifact_not_editable_by_different_user(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        self._check_artifact_patch(
            '/withprops/v1.0/%s' % art['id'],
            data=[{'op': 'replace', 'value': 'public', 'path': '/visibility'}])
        self._set_user('user2')
        self._check_artifact_patch(
            '/withprops/v1.0/%s' % art['id'],
            data=[{'op': 'replace', 'value': 'private',
                   'path': '/visibility'}], status=403)

    def test_public_artifact_editable_by_admin(self):
        data = {'name': 'an artifact',
                'version': '42'}
        art = self._create_artifact('withprops', data=data)
        self._check_artifact_patch(
            '/withprops/v1.0/%s' % art['id'],
            data=[{'op': 'replace', 'value': 'public', 'path': '/visibility'}])
        self._set_user('admin')
        self._check_artifact_patch(
            '/withprops/v1.0/%s' % art['id'],
            data=[{'op': 'replace', 'value': 'private',
                   'path': '/visibility'}], status=200)

    def test_list_artifact_types(self):
        actual = {
            u'artifact_types': [
                {u'displayed_name': u'NoProp',
                 u'type_name': u'NoProp',
                 u'versions':
                     [{u'id': u'v0.5',
                       u'link': u'http://127.0.0.1:%d/v0.1/'
                                u'artifacts/noprop/v0.5'
                                % self.api_port},
                      {u'id': u'v1.0',
                       u'link': u'http://127.0.0.1:%d/v0.1/'
                                u'artifacts/noprop/v1.0'
                                % self.api_port}]},
                {u'displayed_name': u'WithBlob',
                 u'type_name': u'WithBlob',
                 u'versions':
                     [{u'id': u'v1.0',
                       u'link':
                           u'http://127.0.0.1:%d/v0.1/artifacts/withblob/v1.0'
                           % self.api_port}]},
                {u'displayed_name': u'WithProps',
                 u'type_name': u'WithProps',
                 u'versions':
                     [{u'id': u'v1.0',
                       u'link':
                           u'http://127.0.0.1:%d/v0.1/artifacts/withprops/v1.0'
                           % self.api_port}]}]}

        response = self._check_artifact_get("", status=200)
        response[u'artifact_types'].sort(key=lambda x: x[u'type_name'])
        for artifact_type in response[u'artifact_types']:
            artifact_type[u'versions'].sort(key=lambda x: x[u'id'])

        self.assertEqual(actual, response)

    def test_invalid_content_type(self):
        data = {'name': 'name1', 'version': '2.2'}
        self._check_artifact_post('/withprops/v1.0/drafts',
                                  data=data,
                                  status=400,
                                  headers={'Content-Type': 'lalala'})

    def test_filter_by_non_dict_props(self):
        data = {'name': 'art1',
                'version': '4.2',
                'prop2': 12
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art2',
                'version': '4.2',
                'prop2': 10
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art3',
                'version': '4.2',
                'prop2': 10
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art4',
                'version': '4.3',
                'prop2': 33
                }
        self._create_artifact('withprops', data=data)

        result = self._check_artifact_get(
            '/withprops/v1.0/drafts?name=art2')['artifacts']
        self.assertEqual(1, len(result))

        result = self._check_artifact_get(
            '/withprops/v1.0/drafts?prop2=10')['artifacts']
        self.assertEqual(2, len(result))

    def test_filter_by_dict_props(self):
        data = {'name': 'art1',
                'version': '4.2',
                'dict_prop':
                    {'foo': 'Moscow',
                     'bar_list': [42, 44]}
                }
        self._create_artifact('withprops', data=data)
        data = {'name': 'art2',
                'version': '4.2',
                'dict_prop':
                    {'foo': 'Saratov',
                     'bar_list': [42, 42]}
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art3',
                'version': '4.2',
                'dict_prop':
                    {'foo': 'Saratov',
                     'bar_list': [42, 44]}
                }
        self._create_artifact('withprops', data=data)

        url = '/withprops/v1.0/drafts?dict_prop.foo=Saratov'
        result = self._check_artifact_get(url=url)

        self.assertEqual(2, len(result))

        url = '/withprops/v1.0/drafts?dict_prop.bar_list=44'
        result = self._check_artifact_get(url=url)

        self.assertEqual(2, len(result))

    def test_transformation_versions(self):
        data = {'name': 'art1',
                'version': '1'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art2',
                'version': '1.0'}
        art2 = self._create_artifact('noprop', data=data)

        v1 = art1.get("version")
        v2 = art2.get("version")

        self.assertEqual('1.0.0', v1)
        self.assertEqual('1.0.0', v2)

    def test_filter_by_ge_version(self):
        data = {'name': 'art1',
                'version': '4.0.0'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.0.1'}
        art2 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art3 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art4 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0'}
        art5 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '5.0.0'}
        art6 = self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.0.1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art2, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.2.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:5.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art6]
        self.assertEqual(actual, result)

    def test_filter_by_gt_version(self):
        data = {'name': 'art1',
                'version': '4.0.0'}
        self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.0.1'}
        art2 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art3 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art4 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0'}
        art5 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '5.0.0'}
        art6 = self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:4.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art2, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:4.0.1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:4.2.0'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art6]
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=gt:5.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = []
        self.assertEqual(actual, result)

    def test_filter_by_le_version(self):
        data = {'name': 'art1',
                'version': '4.0.0'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.0.1'}
        art2 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art3 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art4 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0'}
        art5 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '5.0.0'}
        art6 = self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art1]
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.0.1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.2.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art5]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:5.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

    def test_filter_by_lt_version(self):
        data = {'name': 'art1',
                'version': '4.0.0'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.0.1'}
        art2 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art3 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art4 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0'}
        art5 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '5.0.0'}
        self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:4.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = []
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:4.0.1'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art1]
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:4.2.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=lt:5.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art5]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

    def test_filter_by_ne_version(self):
        data = {'name': 'art1',
                'version': '4.0.0'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.0.1'}
        art2 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art3 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art4 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0'}
        art5 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '5.0.0'}
        art6 = self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:4.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art2, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:4.0.1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art3, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art4, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art5, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:4.2.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art6]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ne:5.0.0'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2, art3, art4, art5]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

    def test_filter_by_pre_release_version(self):
        data = {'name': 'art1',
                'version': '4.2.0-1'}
        art1 = self._create_artifact('noprop', data=data)

        data = {'name': 'art1',
                'version': '4.2.0-2'}
        art2 = self._create_artifact('noprop', data=data)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art2]
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.2.0-2'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=ge:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        result.sort(key=lambda x: x['id'])

        actual = [art1, art2]
        actual.sort(key=lambda x: x['id'])
        self.assertEqual(actual, result)

        url = '/noprop/v1.0/drafts?name=art1&version=le:4.2.0-1'
        result = self._check_artifact_get(url=url)['artifacts']
        actual = [art1]
        self.assertEqual(actual, result)

    def test_filter_by_range_props(self):
        data = {'name': 'art1',
                'version': '4.2',
                'prop2': 10
                }
        self._create_artifact('withprops', data=data)
        data = {'name': 'art2',
                'version': '4.2',
                'prop2': 100
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art3',
                'version': '4.2',
                'prop2': 1000
                }
        self._create_artifact('withprops', data=data)

        url = '/withprops/v1.0/drafts?prop2=gt:99&prop2=lt:101'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(1, len(result))

        url = '/withprops/v1.0/drafts?prop2=gt:99&prop2=lt:2000'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(2, len(result))

    def test_filter_by_tags(self):
        data = {'name': 'art1',
                'version': '4.2',
                'tags': ['hyhyhy', 'tytyty']
                }
        self._create_artifact('withprops', data=data)
        data = {'name': 'art2',
                'version': '4.2',
                'tags': ['hyhyhy', 'cicici']
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art3',
                'version': '4.2',
                'tags': ['ededed', 'bobobo']
                }
        self._create_artifact('withprops', data=data)

        url = '/withprops/v1.0/drafts?tags=hyhyhy'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(2, len(result))

        url = '/withprops/v1.0/drafts?tags=cicici&tags=hyhyhy'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(1, len(result))

    def test_filter_by_latest_version(self):
        data = {'name': 'art1',
                'version': '1.2',
                'tags': ['hyhyhy', 'tytyty']
                }
        self._create_artifact('withprops', data=data)
        data = {'name': 'latest_artifact',
                'version': '3.2',
                'tags': ['hyhyhy', 'cicici']
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'latest_artifact',
                'version': '3.2',
                'tags': ['ededed', 'bobobo']
                }
        self._create_artifact('withprops', data=data)

        url = '/withprops/v1.0/drafts?version=latest&name=latest_artifact'
        result = self._check_artifact_get(url=url)

        self.assertEqual(2, len(result))

        url = '/withprops/v1.0/drafts?version=latest'
        self._check_artifact_get(url=url, status=400)

    def test_filter_by_version_only(self):
        data = {'name': 'art1',
                'version': '3.2'
                }
        self._create_artifact('withprops', data=data)
        data = {'name': 'art2',
                'version': '4.2'
                }
        self._create_artifact('withprops', data=data)

        data = {'name': 'art3',
                'version': '4.3'
                }
        self._create_artifact('withprops', data=data)

        url = '/withprops/v1.0/drafts?version=gt:4.0&version=lt:10.1'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(2, len(result))

        url = '/withprops/v1.0/drafts?version=gt:4.0&version=ne:4.3'
        result = self._check_artifact_get(url=url)['artifacts']

        self.assertEqual(1, len(result))

    def test_operation_patch_with_blob(self):
        data = {'name': 'art1',
                'version': '3.2'
                }
        art = self._create_artifact('withblob', data=data)

        msg = 'Invalid request PATCH for work with blob'

        result = self._check_artifact_patch(
            '/withblob/v1.0/%s' % art['id'],
            status=400,
            data=[{'op': 'replace',
                   'value': 'public',
                   'path': '/blob1'}])
        self.assertIn(msg, result)

        result = self._check_artifact_patch(
            '/withblob/v1.0/%s' % art['id'],
            status=400,
            data=[{'op': 'remove',
                   'value': 'public',
                   'path': '/blob1'}])
        self.assertIn(msg, result)

        result = self._check_artifact_patch(
            '/withblob/v1.0/%s' % art['id'],
            status=400,
            data=[{'op': 'add',
                   'value': 'public',
                   'path': '/blob1'}])
        self.assertIn(msg, result)

    def test_filter_by_bad_version(self):
        bad_versions = ['kkk', '1.k', 'h.0', '1.3.hf', 's.9.2s2']
        response_string = ('The format of the version %s is not valid. '
                           'Use semver notation')
        for bad_version in bad_versions:
            url = '/withprops/v1.0/drafts?version=gt:%s' % bad_version
            result = self._check_artifact_get(url=url, status=400)
            self.assertIn(response_string % bad_version, result)

    def test_circular_dependency(self):
        data = {'name': 'artifact',
                'version': '12'}
        art = self._create_artifact('withprops', data=data)

        upd = self._check_artifact_post(
            '/withprops/v1.0/%s/depends_on' % art['id'],
            data={'data': art['id']}, status=400)
        self.assertIn(
            'Artifact with a circular dependency can not be created', upd)
